//=============================================================================
// uoTESEvent.js
//=============================================================================

/*:ja
 * @plugindesc TESイベントファイルのロード機能を追加します。
 * @author RaTTiE
 *
 * @param Main Dir
 * @desc TESイベントファイルの格納先です。
 * @default ./data/tes/
 * 
 * @param Patch Dir
 * @desc TESパッチイベントファイルの格納先です。
 * @default ./tespatch/
 * 
 * @param Patch Enable
 * @desc パッチ機能の有効状態です(オフラインモード時のみ有効)。
 * @default false
 *
 * @help
 * 
 * プラグインコマンド:
 *   uoTESEvent link filename label      # TESイベントへリンクする
 *   uoTESEvent release filename         # TESイベントをキャッシュから削除する
 *   uoTESEvent preload filename         # TESイベントを事前ロードする
 */

(function() {
	var parameters = PluginManager.parameters('uoTESEvent');
	var mainDir = String(parameters['Main Dir']) || './data/tes/';
	var patchDir = String(parameters['Patch Dir']) || './tespatch/';
	var patchEnabled = parameters['Patch Enable'] == 'true';
	var $dataTesEvents = {};

	// Interpreterの改変

	/**
	 * プラグインコマンドの実行
	 */
	var orgGame_Interpreter_pluginCommand = Game_Interpreter.prototype.pluginCommand;
	Game_Interpreter.prototype.pluginCommand = function(command, args) {
		orgGame_Interpreter_pluginCommand.call(this, command, args);

		if (command === 'uoTESEvent') {
			switch (args[0]) {
				case 'link':
					this.linkTes(args[1], args[2]);
					break;
				case 'preload':
					this.preloadTes(args[1]);
					break;
				case 'release':
					this.releaseTes(args[1]);
					break;
			}
		}
	};

	/**
	 * コメント文
	 */
	var orgGame_Interpreter_command108 = Game_Interpreter.prototype.command108;
	Game_Interpreter.prototype.command108 = function(){
		var ret = orgGame_Interpreter_command108.call(this);
		if (!ret) return ret;

		this.linkTesFromComment(this._comments);
		return true;
	};

	/**
	 * TESイベントへのリンクを行う
	 */
	Game_Interpreter.prototype.linkTes = function(filename, label) {
		var that = this;

		TesManager.link(filename, function(list) {
			if (list && list.length > 0) {
				var eventId = that.isOnCurrentMap() ? that._eventId : 0;

				that._childInterpreter = new Game_Interpreter(that._depth + 1);
				that._childInterpreter.setup(list, eventId);
				that._childInterpreter.linkLabel(label);
			}
		});
		this.setWaitMode('tes_load');
	};

	/**
	 * TESイベントへのリンクをコメントから行う
	 */
	Game_Interpreter.prototype.linkTesFromComment = function(comments) {
		var regex = /<link\s+(\S+)(\s+(\S+))?>/;

		for (var i = 0; i < comments.length; i++) {
			var match = regex.exec(comments[i]);

			if (match) {
				this.linkTes(match[1], match[3]);
			}
		}
	};

	/**
	 * ラベル行へインデックスを設定する
	 */
	Game_Interpreter.prototype.linkLabel = function(label) {
		if (!label) return;

		for (var i = 0; i < this._list.length; i++) {
			var event = this._list[i];
			if (event.code == 118 && event.parameters[0] == label) {
				this._index = i;
				return;
			}
		}

		throw new Error('The link label is not found');
	}

	/**
	 * TESイベントのプリロードを行う
	 */
	Game_Interpreter.prototype.preloadTes = function(filename) {
		this.setWaitMode('tes_load');

		TesManager.link(filename, function(list) { });
	};

	/**
	 * TESイベントをキャッシュから削除する
	 */
	Game_Interpreter.prototype.releaseTes = function(filename) {
		TesManager.release(filename);
	};

	/**
	 * ウェイトモードを更新する
	 */
	var orgGame_Interpreter_updateWaitMode = Game_Interpreter.prototype.updateWaitMode;
	Game_Interpreter.prototype.updateWaitMode = function() {
		var waiting = false;

		switch (this._waitMode) {
			case 'tes_load':
				waiting = !TesManager.isEventLoaded();

				break;
			default:
				return orgGame_Interpreter_updateWaitMode.call(this);
		}

		if (!waiting) {
			this._waitMode = '';
		}

		return waiting;
	};

	/**
	 * TES管理クラス
	 */
	function TesManager() {
		throw new Error('This is a static class');
	}
	TesManager._errorUrl = null;

	/**
	 * TESのロードを行う
	 */
	TesManager.load = function(dir, filename, onLoad) {
		var xhr = new XMLHttpRequest();
		var url = dir + filename + '.json';

		xhr.open('GET', url);
		xhr.overrideMimeType('application/json');
		xhr.onload = function() {
			if (xhr.status < 400) {
				var list = JSON.parse(xhr.responseText);

				$dataTesEvents[filename] = list;
				onLoad(list);
			}
		};
		xhr.onerror = function() {
			TesManager._errorUrl = TesManager._errorUrl || url;
		}
		$dataTesEvents[filename] = null;
		xhr.send();
	};

	/**
	 * TESのリンクを行う
	 */
	TesManager.link = function(filename, onLoad) {
		// キャッシュがあればそのまま返す
		if ($dataTesEvents[filename]) {
			onLoad($dataTesEvents[filename]);

			return;
		}

		var dir = mainDir;

		// パッチ有効時
		if (patchEnabled) {
			// パッチはオフライン時のみ有効
			if (typeof require === 'function' && typeof process === 'object') {
				var fs = require('fs');
				var path = StorageManager.localFileDirectoryPath();
				path = path.substring(0, path.length - 5);

				var dirs = fs.readdirSync(path + patchDir);

				var patches = [];
				var regex = /^patch(\d+)/;

				// パッチフォルダ直下のフォルダを走査
				dirs = dirs.filter(function(localDir) {
					return fs.statSync(path + patchDir + localDir).isDirectory();
				});

				// 所定形式のフォルダのみ取得
				for (var i = 0; i < dirs.length; i++) {
					var match = regex.exec(dirs[i]);

					if (match) {
						patches.push(parseInt(match[1], 10));
					}
				}

				// 降順ソート
				patches.sort(function(a, b) { return a > b ? -1 : a < b ? 1 : 0; });

				// パッチフォルダ内に指定したファイルがあればパッチフォルダからロード
				for (var i = 0; i < patches.length; i++) {
					var localPatchDir = path + patchDir + 'patch' + patches[i].toString() + '/';

					if (fs.existsSync(localPatchDir + filename + '.json')) {
						dir = localPatchDir;

						var list = JSON.parse(fs.readFileSync(dir + filename + '.json', 'utf8'));
						$dataTesEvents[filename] = list;
						onLoad(list);

						return;
					}
				}
			}
		}

		TesManager.load(dir, filename, onLoad);
	};

	/**
	 * TESイベントをキャッシュから削除する
	 */
	TesManager.release = function(filename) {
		// ファイル名省略時は全削除
		if (!filename) {
			$dataTesEvents = {};

			return;
		}

		if ($dataTesEvents[filename]) delete $dataTesEvents[filename];
	};

	/**
	 * イベントデータがロード済か？
	 */
	TesManager.isEventLoaded = function() {
		TesManager.checkError();

		for (var key in $dataTesEvents) {
			if (!$dataTesEvents[key]) {
				return false;
			}
		}

		return true;
	};

	/**
	 * エラーチェックを行う
	 */
	TesManager.checkError = function() {
		if (TesManager._errorUrl) {
			throw new Error('Failed to load: ' + TesManager._errorUrl);
		}
	}
})();
